package cn.com.qmth.stmms.admin.thread;

import cn.com.qmth.stmms.biz.exam.model.*;
import cn.com.qmth.stmms.biz.exam.service.*;
import cn.com.qmth.stmms.biz.lock.LockService;
import cn.com.qmth.stmms.biz.mark.service.MarkService;
import cn.com.qmth.stmms.biz.report.service.ReportService;
import cn.com.qmth.stmms.biz.report.utils.ReportContext;
import cn.com.qmth.stmms.biz.utils.ScoreCalculateUtil;
import cn.com.qmth.stmms.biz.utils.ScoreInfo;
import cn.com.qmth.stmms.common.enums.ExamStatus;
import cn.com.qmth.stmms.common.enums.LockType;
import org.apache.commons.lang.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.*;

public class ScoreCalculateThread implements Runnable {

    protected static Logger log = LoggerFactory.getLogger(ScoreCalculateThread.class);

    private ExamStudentService studentService;

    private ExamQuestionService questionService;

    private MarkService markService;

    private int examId;

    private Set<String> subjectCodeSet;

    private LockService lockService;

    private ReportService reportService;

    private Map<String, List<ExamQuestion>> objectiveMap;

    private Map<String, List<ExamQuestion>> subjectiveMap;

    private Map<String, ExamSubject> subjectMap;

    private Map<String, List<MarkGroup>> groupMap;

    private MarkGroupService groupService;

    private ExamSubjectService subjectService;

    private ExamService examService;

    private ReportContext context;

    public ScoreCalculateThread(int examId, Set<String> subjectCodeSet, LockService lockService,
            ExamStudentService studentService, ExamQuestionService questionService, MarkService markService,
            ReportService reportService, ExamService examService, ExamSubjectService subjectService,
            MarkGroupService groupService) {
        this.examId = examId;
        this.subjectCodeSet = subjectCodeSet != null ? subjectCodeSet : new HashSet<>();
        this.lockService = lockService;
        this.studentService = studentService;
        this.questionService = questionService;
        this.markService = markService;
        this.reportService = reportService;
        this.examService = examService;
        this.subjectService = subjectService;
        this.groupService = groupService;
        this.objectiveMap = new HashMap<>();
        this.subjectiveMap = new HashMap<>();
        this.subjectMap = new HashMap<>();
        this.groupMap = new HashMap<>();
    }

    @Override
    public void run() {
        // 获取考试信息并判断状态
        Exam exam = examService.findById(examId);
        if (exam == null || exam.getStatus() != ExamStatus.START) {
            log.info("calculate exception for examId=" + examId + ", exam is null or status error");
            return;
        }
        log.info("start calculate for examId=" + examId + ", subjectCode count=" + subjectCodeSet.size());
        for (String subjectCode : subjectCodeSet) {
            // 尝试上锁，失败直接跳过
            if (!lockService.trylock(LockType.SCORE_CALCULATE, examId, subjectCode)) {
                log.warn("calculate locked for examId=" + examId + ", subjectCode=" + subjectCode);
                continue;
            }
            try {
                log.info("start calculate for examId=" + examId + ", subjectCode=" + subjectCode);
                // 删除原有统计数据
                reportService.deleteData(examId, subjectCode);
                context = new ReportContext(exam);
                long count = 0;
                long totalCount = studentService.countByExamIdAndSubjectCode(examId, subjectCode);
                int pageNumber = 1;
                int pageSize = 1000;
                List<ExamStudent> list = studentService
                        .findByExamIdAndSubjectCode(examId, subjectCode, pageNumber, pageSize);
                while (list != null && list.size() > 0) {
                    for (ExamStudent student : list) {
                        // 先统分
                        calculate(student);
                        // 后统计
                        statistic(student);
                        // 累加计数
                        count++;
                    }
                    // 修正进度百分比
                    double progress = count * 100.0 / totalCount;
                    if (progress == 100 && count < totalCount) {
                        progress = 99;
                    } else if (progress == 0 && count > 0) {
                        progress = 1;
                    }
                    // 更新进度百分比
                    subjectService.updateCalculateProgress(examId, subjectCode, progress);
                    pageNumber++;
                    list = studentService.findByExamId(examId, pageNumber, pageSize);
                }
                // 结束统计
                context.save();
                // 统分结束修改标记
                subjectService.updateNeedCalculate(examId, subjectCode, false);
            } catch (Exception e) {
                log.error("calculate exception for examId=" + examId + ", subjectCode=" + subjectCode, e);
            } finally {
                lockService.unlock(LockType.SCORE_CALCULATE, examId, subjectCode);
                log.info("finish calculate for examId=" + examId + ", subjectCode=" + subjectCode);
            }
        }
    }

    private void calculate(ExamStudent student) {
        // 未上传、缺考、违纪的考生不统分
        if (!student.isUpload() || student.isAbsent() || student.isBreach()) {
            return;
        }
        try {
            ScoreCalculateUtil util = ScoreCalculateUtil.instance(student);
            ScoreInfo info = util
                    .calculate(findQuestionList(student.getSubjectCode(), student.getPaperType(), true), null);

            student.setObjectiveScore(info.getObjectiveScore());
            student.setScoreList(info.getScoreList(), true);
            // 更新客观题得分
            studentService.updateObjectiveScore(student.getId(), student.getObjectiveScore(),
                    student.getObjectiveScoreList());
            // 增加主观题总分统计
            markService.scoreCalculate(student, findMarkGroup(student.getSubjectCode()));
        } catch (Exception e) {
            log.error("calculate error for studentId=" + student.getId(), e);
        }
    }

    private void statistic(ExamStudent student) {
        student.setSubject(findExamSubject(student.getSubjectCode()));
        student.setObjectiveQuestionList(findQuestionList(student.getSubjectCode(), student.getPaperType(), true));
        student.setSubjectiveQuestionList(findQuestionList(student.getSubjectCode(), student.getPaperType(), false));
        context.process(student);
    }

    private List<ExamQuestion> findQuestionList(String subjectCode, String paperType, boolean objective) {
        if (objective) {
            String key = subjectCode + "_" + StringUtils.trimToEmpty(paperType);
            List<ExamQuestion> list = objectiveMap.get(key);
            if (list == null) {
                list = questionService
                        .findByExamAndSubjectAndObjectiveAndPaperType(examId, subjectCode, true, paperType);
                objectiveMap.put(key, list);
            }
            return list;
        } else {
            List<ExamQuestion> list = subjectiveMap.get(subjectCode);
            if (list == null) {
                list = questionService.findByExamAndSubjectAndObjective(examId, subjectCode, false);
                subjectiveMap.put(subjectCode, list);
            }
            return list;
        }
    }

    private ExamSubject findExamSubject(String subjectCode) {
        ExamSubject subject = subjectMap.get(subjectCode);
        if (subject == null) {
            subject = subjectService.find(examId, subjectCode);
            subjectMap.put(subjectCode, subject);
        }
        return subject;
    }

    private List<MarkGroup> findMarkGroup(String subjectCode) {
        List<MarkGroup> list = groupMap.get(subjectCode);
        if (list == null) {
            list = groupService.findByExamAndSubject(examId, subjectCode);
            groupMap.put(subjectCode, list);
        }
        return list;
    }
}
